@@ -20,6 +20,7 @@ import sys  # for processing of command line args
 import threading
 import tempfile
 import signal
+import mimetypes
 import pathlib
 import platform
 import configparser
@@ -30,6 +31,13 @@ import gettext
 import gc
 from urllib.request import url2pathname
 
+try:
+    import img2pdf
+    img2pdf.Image.init()
+    img2pdf_supported_img = [i for i in img2pdf.Image.MIME.values() if i.split('/')[0] == 'image']
+except ImportError:
+    img2pdf = None
+
 sharedir = os.path.join(sys.prefix, 'share')
 basedir = '.'
 if getattr(sys, 'frozen', False):
@@ -65,10 +73,15 @@ else:
         pass
 
 APPNAME = 'PDF Arranger'
-VERSION = '1.5.3'
+VERSION = '1.6.0'
 WEBSITE = 'https://github.com/jeromerobert/pdfarranger'
 LICENSE = 'GNU General Public License (GPL) Version 3.'
 
+# Add support for dnd to other instance and insert file at drop location in Windows
+if os.name == 'nt':
+    import keyboard  # to get control key state when drag to other instance
+    os.environ['GDK_WIN32_USE_EXPERIMENTAL_OLE2_DND'] = 'true'
+
 import gi
 
 # check that we don't need GObject.threads_init()
@@ -186,7 +199,8 @@ def warn_dialog(func):
 
         def __call__(self, message, category, filename, lineno, f=None, line=None):
             s = warnings.formatwarning(message, category, filename, lineno, line)
-            sys.stderr.write(s + '\n')
+            if sys.stderr is not None:
+                sys.stderr.write(s + '\n')
             self.buffer += str(message) + '\n'
 
     def wrapper(*args, **kwargs):
@@ -203,7 +217,7 @@ def warn_dialog(func):
     return wrapper
 
 
-def get_file_path_from_dnd_dropped_uri(uri):
+def get_file_path_from_uri(uri):
     """Extracts the path from an uri"""
     uri = uri[5:]  # remove 'file:'
     path = url2pathname(uri)  # escape special chars
@@ -261,17 +275,21 @@ class PdfArranger(Gtk.Application):
         self.nfile = 0
         self.iv_auto_scroll_direction = 0
         self.iv_auto_scroll_timer = None
+        self.vp_css_margin = 0
         self.pdfqueue = []
         self.metadata = {}
         self.pressed_button = None
         self.rendering_thread = None
         self.export_file = None
+        self.drag_path = None
+        self.drag_pos = Gtk.IconViewDropPosition.DROP_RIGHT
 
         # Clipboard for cut copy paste
-        if os.name == 'nt':
-            self.clipboard = Gtk.Clipboard.get(Gdk.SELECTION_CLIPBOARD)
-        else:
-            self.clipboard = Gtk.Clipboard.get(Gdk.Atom.intern('_SELECTION_PDFARRANGER', False))
+        self.clipboard_default = Gtk.Clipboard.get(Gdk.SELECTION_CLIPBOARD)
+        if os.name == 'posix':
+            # "private" clipboard does not work in Windows so we can't use it here
+            self.clipboard_pdfarranger = Gtk.Clipboard.get(Gdk.Atom.intern('_SELECTION_PDFARRANGER',
+                                                                           False))
 
     def do_open(self, files, _n, _hints):
         """ https://lazka.github.io/pgi-docs/Gio-2.0/classes/Application.html#Gio.Application.do_open """
@@ -362,6 +380,7 @@ class PdfArranger(Gtk.Application):
             self.set_accels_for_action("win." + a, [k] if isinstance(k, str) else k)
         # Disable actions
         self.iv_selection_changed_event()
+        self.__update_num_pages(self.iconview.get_model())
         self.undomanager.set_actions(self.window.lookup_action('undo'),
                                      self.window.lookup_action('redo'))
 
@@ -384,20 +403,17 @@ class PdfArranger(Gtk.Application):
         if self.config.maximized():
             self.window.maximize()
         self.window.set_default_size(*self.config.window_size())
-        self.window.connect('delete_event', self.close_application)
+        self.window.connect('delete_event', self.on_quit)
 
         if hasattr(GLib, "unix_signal_add"):
             GLib.unix_signal_add(GLib.PRIORITY_DEFAULT, signal.SIGINT, self.close_application)
 
         # Create a scrolled window to hold the thumbnails-container
         self.sw = self.uiXML.get_object('scrolledwindow')
-        self.sw.drag_dest_set(Gtk.DestDefaults.MOTION |
-                              Gtk.DestDefaults.HIGHLIGHT |
-                              Gtk.DestDefaults.DROP |
-                              Gtk.DestDefaults.MOTION,
+        self.sw.drag_dest_set(Gtk.DestDefaults.HIGHLIGHT |
+                              Gtk.DestDefaults.DROP,
                               self.TARGETS_SW,
-                              Gdk.DragAction.COPY |
-                              Gdk.DragAction.MOVE)
+                              Gdk.DragAction.COPY)
         self.sw.connect('drag_data_received', self.sw_dnd_received_data)
         self.sw.connect('button_press_event', self.sw_button_press_event)
         self.sw.connect('scroll_event', self.sw_scroll_event)
@@ -440,7 +456,8 @@ class PdfArranger(Gtk.Application):
                                                Gdk.DragAction.COPY |
                                                Gdk.DragAction.MOVE)
         self.iconview.enable_model_drag_dest(self.TARGETS_IV,
-                                             Gdk.DragAction.DEFAULT)
+                                             Gdk.DragAction.COPY |
+                                             Gdk.DragAction.MOVE)
         self.iconview.connect('drag_begin', self.iv_drag_begin)
         self.iconview.connect('drag_data_get', self.iv_dnd_get_data)
         self.iconview.connect('drag_data_received', self.iv_dnd_received_data)
@@ -594,26 +611,41 @@ class PdfArranger(Gtk.Application):
            iconview cols no."""
         if len(self.model) > 0:
             # scale*page_width*(1-crop_left-crop_right)
-            item_width = int(max(row[4] * row[11] * (1. - row[7] - row[8])
+            item_width = int(max(0.5 + int(row[4] * row[11]) * (1. - row[7] - row[8])
                                  for row in self.model))
             item_padding = self.iconview.get_item_padding()
             cellthmb_xpad, _cellthmb_ypad = self.cellthmb.get_padding()
-            min_col_spacing = 7 + 2 * (item_padding + cellthmb_xpad)
-            min_margin = 14
+            border_and_shadow = 7  # 2*th1+th2 set in iconview.py
+            # cell width min limit 50 is set in gtkiconview.c
+            cell_width = max(item_width + 2 * cellthmb_xpad + border_and_shadow, 50)
+            self.cellthmb.set_fixed_size(cell_width, -1)
+            padded_cell_width = cell_width + 2 * item_padding
+            min_col_spacing = 5
+            min_margin = 11
             iw_width = window.get_size()[0]
-            # 2 * min_margin + col_num * item_width
+            # 2 * min_margin + col_num * padded_cell_width
             #  + min_col_spacing * (col_num+1) = iw_width
-            # min_margin+margin = min_col_spacing+col_spacing
-            #  = (iw_width - col_num * item_width) / (col_num+1)
             col_num = (iw_width - 2 * min_margin - min_col_spacing) //\
-                      (item_width + min_col_spacing)
-            spacing = (iw_width - col_num * item_width) // (col_num + 1)
+                      (padded_cell_width + min_col_spacing)
+            spacing = (iw_width - col_num * padded_cell_width - 2 * min_margin) // (col_num + 1)
+            margin = (iw_width - col_num * (padded_cell_width + spacing) + spacing) // 2
             if col_num == 0:
                 col_num = 1
-                spacing = min_margin
+                margin = 6
             self.iconview.set_columns(col_num)
-            self.iconview.set_column_spacing(spacing - min_col_spacing)
-            self.iconview.set_margin_left(spacing - min_margin)
+            self.iconview.set_column_spacing(spacing)
+            self.iconview.set_margin(margin)
+            if self.vp_css_margin != 6 - margin:
+                # remove margin on top and bottom
+                self.vp_css_margin = 6 - margin
+                css_data = 'viewport {margin-top:' + str(self.vp_css_margin) + 'px;\
+                margin-bottom:' + str(self.vp_css_margin) + 'px;}'
+                style_provider = Gtk.CssProvider()
+                style_provider.load_from_data(bytes(css_data.encode()))
+                Gtk.StyleContext.add_provider_for_screen(
+                    Gdk.Screen.get_default(),
+                    style_provider,
+                    Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION)
 
     def update_geometry(self, treeiter):
         """Recomputes the width and height of the rotated page and saves
@@ -635,7 +667,28 @@ class PdfArranger(Gtk.Application):
 
         self.model.set(treeiter, 11, w1, 12, h1)
 
-    def on_quit(self, _action, _param, _unknown):
+    def on_quit(self, _action, _param=None, _unknown=None):
+        if self.is_unsaved:
+            if self.export_file:
+                msg = _('Save changes to “{}” before closing?').format(os.path.basename(self.export_file))
+            else:
+                msg = _('Save changes before closing?')
+            d = Gtk.MessageDialog(self.window, 0, Gtk.MessageType.WARNING, Gtk.ButtonsType.NONE, msg)
+            d.format_secondary_markup(_("Your changes will be lost if you don’t save them."))
+            d.add_buttons(_('Do_n’t Save'), 1, _('_Cancel'), 2, _('_Save'), 3)
+            response = d.run()
+            d.destroy()
+
+            if response == 2:
+                # Returning True to stop self.window delete_event propagation.
+                return True
+            elif response == 3:
+                # Save.
+                self.save_or_choose()
+                # Quit only if it has been really saved.
+                if self.is_unsaved:
+                    return True
+
         self.close_application()
 
     def close_application(self, _widget=None, _event=None, _data=None):
@@ -707,6 +760,11 @@ class PdfArranger(Gtk.Application):
         return all_files
 
     def on_action_save(self, _action, _param, _unknown):
+        self.save_or_choose()
+
+    def save_or_choose(self):
+        """Saves to the previously exported file or shows the export dialog if
+        there was none."""
         try:
             if self.export_file:
                 self.save(False, self.export_file)
@@ -757,6 +815,15 @@ class PdfArranger(Gtk.Application):
         filter_all.add_pattern('*')
         chooser.add_filter(filter_all)
 
+        if img2pdf:
+            filter_image = Gtk.FileFilter()
+            filter_image.set_name(_('Supported image files'))
+            for mime in img2pdf_supported_img:
+                filter_image.add_mime_type(mime)
+                for extension in mimetypes.guess_all_extensions(mime):
+                    filter_image.add_pattern('*' + extension)
+            chooser.add_filter(filter_image)
+
         filter_pdf = Gtk.FileFilter()
         filter_pdf.set_name(_('PDF files'))
         filter_pdf.add_pattern('*.pdf')
@@ -815,13 +882,41 @@ class PdfArranger(Gtk.Application):
 
     @staticmethod
     def data_to_pageadder(data, pageadder):
-        """Data to pageadder"""
-
+        """Data to pageadder."""
         tmp = data.pop(0).split('\n')
         filename = tmp[0]
-        npage, angle = [int(k) for k in tmp[1:3]]
-        crop = [float(side) for side in tmp[3:7]]
-        pageadder.addpages(filename, npage, angle, crop)
+        npage = int(tmp[1])
+        if len(tmp) < 3:  # Only when paste files interleaved
+            pageadder.addpages(filename, npage)
+        else:
+            angle = int(tmp[2])
+            crop = [float(side) for side in tmp[3:7]]
+            pageadder.addpages(filename, npage, angle, crop)
+
+    def is_data_valid(self, data):
+        """Validate data to be pasted from clipboard. Only used in Windows."""
+        data_copy = data.copy()
+        data_valid = True
+        while data_copy:
+            try:
+                tmp = data_copy.pop(0).split('\n')
+                filename = tmp[0]
+                npage = int(tmp[1])
+                angle = int(tmp[2])
+                crop = [float(side) for side in tmp[3:7]]
+                if not (os.path.isfile(filename) and len(tmp) == 7 and
+                        npage > 0 and angle in [0, 90, 180, 270] and
+                        all((cr >= 0.0 and cr <= 0.99) for cr in crop) and
+                        (crop[0] + crop[1] <= 0.99) and (crop[2] + crop[3] <= 0.99)):
+                    data_valid = False
+                    break
+            except (ValueError, IndexError):
+                data_valid = False
+                break
+        if not data_valid:
+            message = _('Pasted data not valid. Aborting paste.')
+            self.error_message_dialog(message)
+        return data_valid
 
     def paste_pages(self, data, before, ref_to, select_added):
         """Paste pages to iconview"""
@@ -829,19 +924,24 @@ class PdfArranger(Gtk.Application):
         pageadder = PageAdder(self)
         if ref_to:
             pageadder.move(ref_to, before)
-
-        if not before:
+        if not before and ref_to:
             data.reverse()
+
         while data:
-            try:  # Clipboard can contain not expected data
-                self.data_to_pageadder(data, pageadder)
-            except (ValueError, FileNotFoundError):
-                return False
+            self.data_to_pageadder(data, pageadder)
         return pageadder.commit(select_added, add_to_undomanager=True)
 
-    def paste_pages_interleave(self, data, before, ref_to):
-        """Paste pages interleved to iconview"""
+    def paste_files(self, filepaths, before, ref_to):
+        """Paste files to iconview."""
+        pageadder = PageAdder(self)
 
+        for filepath in filepaths:
+            pageadder.move(ref_to, before)
+            pageadder.addpages(filepath)
+        pageadder.commit(select_added=False, add_to_undomanager=True)
+
+    def paste_pages_interleave(self, data, before, ref_to):
+        """Paste pages or files interleved to iconview."""
         pageadder = PageAdder(self)
         model = self.iconview.get_model()
         iter_to = None
@@ -850,10 +950,7 @@ class PdfArranger(Gtk.Application):
         self.set_unsaved(True)
 
         while data:
-            try:  # Clipboard can contain not expected data
-                self.data_to_pageadder(data, pageadder)
-            except (ValueError, FileNotFoundError):
-                return
+            self.data_to_pageadder(data, pageadder)
 
             pageadder.move(ref_to, before)
             pageadder.commit(select_added=False, add_to_undomanager=False)
@@ -876,58 +973,140 @@ class PdfArranger(Gtk.Application):
         self.clear_selected()
 
     def on_action_cut(self, _action, _param, _unknown):
-        """Cut selected pages to clipboard"""
-
+        """Cut selected pages to clipboard."""
         data = self.copy_pages()
-        self.clipboard.set_text(data, -1)
+        if os.name == 'posix':
+            self.clipboard_pdfarranger.set_text(data, -1)
+            self.clipboard_default.set_text('', -1)
+        if os.name == 'nt':
+            self.clipboard_default.set_text('pdfarranger-clipboard\n' + data, -1)
 
         self.clear_selected()
 
     def on_action_copy(self, _action, _param, _unknown):
-        """Copy selected pages to clipboard"""
-
+        """Copy selected pages to clipboard."""
         data = self.copy_pages()
-        self.clipboard.set_text(data, -1)
+        if os.name == 'posix':
+            self.clipboard_pdfarranger.set_text(data, -1)
+            self.clipboard_default.set_text('', -1)
+        if os.name == 'nt':
+            self.clipboard_default.set_text('pdfarranger-clipboard\n' + data, -1)
 
     def on_action_paste(self, _action, mode, _unknown):
-        """Paste pages from clipboard"""
+        """Paste pages or files from clipboard."""
+        data, data_is_filepaths = self.read_from_clipboard()
+        if not data:
+            return
 
+        pastemodes = {0: 'AFTER', 1: 'BEFORE', 2: 'ODD', 3: 'EVEN'}
+        pastemode = pastemodes[mode.get_int32()]
+
+        ref_to, before = self.set_paste_location(pastemode, data_is_filepaths)
+
+        if pastemode in ['AFTER', 'BEFORE']:
+            if data_is_filepaths:
+                self.paste_files(data, before, ref_to)
+            else:
+                self.paste_pages(data, before, ref_to, select_added=False)
+        elif pastemode in ['ODD', 'EVEN']:
+            if data_is_filepaths:
+                # Generate data to send to paste_pages_interleave
+                filepaths = []
+                try:
+                    for filepath in data:
+                        filemime = mimetypes.guess_type(filepath)[0]
+                        if not filemime:
+                            raise PDFDocError(filepath + ':\n' + _('Unknown file format'))
+                        if filemime == 'application/pdf':
+                            num_pages = exporter.num_pages(filepath)
+                            if num_pages is None:
+                                raise PDFDocError(filepath + ':\n' + _('PDF document is damaged'))
+                            for page in range(1, num_pages + 1):
+                                filepaths.append('\n'.join([filepath, str(page)]))
+                        elif filemime.split('/')[0] == 'image':
+                            filepaths.append('\n'.join([filepath, str(1)]))
+                        else:
+                            raise PDFDocError(filepath + ':\n' + _('File is neither pdf nor image'))
+                except PDFDocError as e:
+                    print(e.message, file=sys.stderr)
+                    self.error_message_dialog(e.message)
+                    return
+                data = filepaths
+            self.paste_pages_interleave(data, before, ref_to)
+
+    def read_from_clipboard(self):
+        """Read data from clipboards. Check if data is copied pages or files."""
+        # In Linux, if default clipboard holds path to pdf or image files,
+        # these files will be pasted with precedence over copied pages.
+        # In Windows default clipboard is used for both copied pages and copied files.
+        # If id "pdfarranger-clipboard" is found pages is expected to be in clipboard, else files.
+        data = self.clipboard_default.wait_for_text()
+        if not data:
+            data = ''
+
+        data_is_filepaths = False
+        if os.name == 'posix' and not data:
+            data = self.clipboard_pdfarranger.wait_for_text()
+            if data:
+                data = data.split('\n;\n')
+        elif os.name == 'nt' and data.startswith('pdfarranger-clipboard\n'):
+            data = data.replace('pdfarranger-clipboard\n', '', 1)
+            data = data.split('\n;\n')
+            if not self.is_data_valid(data):
+                data = []
+        else:
+            data_is_filepaths = True
+            if os.name == 'posix' and data.startswith('x-special/nautilus-clipboard\ncopy'):
+                data = data.replace('x-special/nautilus-clipboard\ncopy', '', 1)
+            rows = data.split('\n')
+            rows = filter(None, rows)
+            data = []
+            for row in rows:
+                if os.name == 'posix' and row.startswith('file:///'):  # Dolphin, Nautilus
+                    row = get_file_path_from_uri(row)
+                elif os.name == 'nt' and row.startswith('"') and row.endswith('"'):
+                    row = row[1:-1]
+                if os.path.isfile(row):
+                    data.append(row)
+                else:
+                    data = []
+                    break
+
+        return data, data_is_filepaths
+
+    def set_paste_location(self, pastemode, data_is_filepaths):
+        """Sets reference where pages should be pasted and if before or after that."""
         model = self.iconview.get_model()
 
         selection = self.iconview.get_selected_items()
         selection.sort(key=lambda x: x.get_indices()[0])
-
-        # mode = 0 paste after
-        # mode = 1 paste before
-        # mode = 2 paste interleave odd
-        # mode = 3 paste interveave even
-        mode = mode.get_int32()
         if len(model) == 0:
             before = True
             ref_to = None
-        elif mode == 0:
-            before = False
-            if len(selection) == 0:
+        elif pastemode == 'AFTER':
+            last_row = model[-1]
+            if len(selection) == 0 or selection[-1] == last_row.path:
+                before = False
                 ref_to = None
+            elif data_is_filepaths:
+                before = True
+                path = selection[-1]
+                iter_next = model.iter_next(model.get_iter(path))
+                path_next = model.get_path(iter_next)
+                ref_to = Gtk.TreeRowReference.new(model, path_next)
             else:
+                before = False
                 ref_to = Gtk.TreeRowReference.new(model, selection[-1])
         else:
-            if mode == 3:
+            if pastemode == 'EVEN':
                 before = False
-            else:
+            else:  # BEFORE or ODD
                 before = True
             if len(selection) == 0:
                 ref_to = Gtk.TreeRowReference.new(model, Gtk.TreePath(0))
             else:
                 ref_to = Gtk.TreeRowReference.new(model, selection[0])
-
-        data = self.clipboard.wait_for_text()
-        if data:
-            data = data.split('\n;\n')
-            if mode == 0 or mode == 1:
-                self.paste_pages(data, before, ref_to, select_added=False)
-            elif mode == 2 or mode == 3:
-                self.paste_pages_interleave(data, before, ref_to)
+        return ref_to, before
 
     @staticmethod
     def iv_drag_begin(iconview, context):
@@ -935,7 +1114,7 @@ class PdfArranger(Gtk.Application):
         selected_count = len(iconview.get_selected_items())
         stock_icon = "gtk-dnd-multiple" if selected_count > 1 else "gtk-dnd"
         iconview.stop_emission('drag_begin')
-        Gtk.drag_set_icon_name(context, stock_icon, 0, 0)
+        Gtk.drag_set_icon_name(context, stock_icon, 16, 16)
 
     def iv_dnd_get_data(self, _iconview, _context,
                         selection_data, _target_id, _etime):
@@ -958,7 +1137,7 @@ class PdfArranger(Gtk.Application):
             return
         selection_data.set(selection_data.get_target(), 8, data.encode())
 
-    def iv_dnd_received_data(self, iconview, context, x, y,
+    def iv_dnd_received_data(self, iconview, context, _x, _y,
                              selection_data, _target_id, etime):
         """Handles received data by drag and drop in iconview"""
 
@@ -967,21 +1146,14 @@ class PdfArranger(Gtk.Application):
         if not data:
             return
         data = data.decode().split('\n;\n')
-        item = iconview.get_dest_item_at_pos(x, y)
-        if item:
-            path, position = item
-            ref_to = Gtk.TreeRowReference.new(model, path)
+        if self.drag_path and len(model) > 0:
+            ref_to = Gtk.TreeRowReference.new(model, self.drag_path)
         else:
             ref_to = None
-            position = Gtk.IconViewDropPosition.DROP_RIGHT
-            if len(model) > 0:  # find the iterator of the last row
-                row = model[-1]
-                ref_to = Gtk.TreeRowReference.new(model, row.path)
-        before = (position == Gtk.IconViewDropPosition.DROP_LEFT
-                  or position == Gtk.IconViewDropPosition.DROP_ABOVE)
+        before = self.drag_pos == Gtk.IconViewDropPosition.DROP_LEFT
         target = selection_data.get_target().name()
         if target == 'MODEL_ROW_INTERN':
-            move = context.get_actions() & Gdk.DragAction.MOVE
+            move = context.get_selected_action() & Gdk.DragAction.MOVE
             self.undomanager.commit("Move" if move else "Copy")
             self.set_unsaved(True)
             data.sort(key=int, reverse=not before)
@@ -1001,13 +1173,8 @@ class PdfArranger(Gtk.Application):
                     model.remove(model.get_iter(ref_from.get_path()))
 
         elif target == 'MODEL_ROW_EXTERN':
-            if not item and self.is_between_items(iconview, x, y):
-                context.finish(False, False, etime)
-                return
-            if not ref_to:
-                data.reverse()
             changed = self.paste_pages(data, before, ref_to, select_added=True)
-            if changed and context.get_actions() & Gdk.DragAction.MOVE:
+            if changed and context.get_selected_action() & Gdk.DragAction.MOVE:
                 context.finish(True, True, etime)
 
     def iv_dnd_data_delete(self, _widget, _context):
@@ -1015,8 +1182,8 @@ class PdfArranger(Gtk.Application):
         been moved to another instance."""
         if self.target_is_intern and os.name == 'nt':
             # Workaround for windows
-            # On Windows this method is triggered even for drag & drop within the same
-            # pdfarranger instance
+            # On Windows this method is in some situations triggered even for drag & drop
+            # within the same pdfarranger instance
             return
         selection = self.iconview.get_selected_items()
         self.undomanager.commit("Move")
@@ -1027,40 +1194,18 @@ class PdfArranger(Gtk.Application):
             path = ref_del.get_path()
             model.remove(model.get_iter(path))
 
-    def is_between_items(self, iconview, x, y):
-        """Find out if drag location is between items."""
-        model = iconview.get_model()
-        if len(model) == 0:
-            return False
-        last_row = model[-1]
-        _x, _y, w, _h = self.cellthmb.do_get_size(iconview)
-        x_step = w
-        y_step = iconview.get_row_spacing() + 2 * iconview.get_item_padding()
-        xy_test = [(x - x_step, y),             # left
-                   (x + x_step, y),             # right
-                   (x, y + y_step),             # down
-                   (x - x_step, y + y_step),    # left-down
-                   (x + x_step, y + y_step)]    # right-down
-
-        for x_t, y_t in xy_test:
-            if x_t < 0:
-                x_t = 0
-            path = iconview.get_path_at_pos(x_t, y_t)
-            if path and not (path == last_row.path and x_t < x):
-                return True
-        return False
-
-    def iv_dnd_motion(self, iconview, _context, x, y, _etime):
-        """Handles auto-scroll when drag up/down. Also reject drop to location between items."""
+    def iv_dnd_motion(self, iconview, context, x, y, etime):
+        """Handles drag motion: autoscroll, select move or copy, select drag cursor location."""
+        # Auto-scroll when drag up/down
         autoscroll_area = 40
         sw_vadj = self.sw.get_vadjustment()
         sw_height = self.sw.get_allocation().height
-        if y - sw_vadj.get_value() < autoscroll_area:
+        if y - sw_vadj.get_value() < autoscroll_area - self.vp_css_margin:
             if not self.iv_auto_scroll_timer:
                 self.iv_auto_scroll_direction = Gtk.DirectionType.UP
                 self.iv_auto_scroll_timer = GObject.timeout_add(150,
                                                                 self.iv_auto_scroll)
-        elif y - sw_vadj.get_value() > sw_height - autoscroll_area:
+        elif y - sw_vadj.get_value() > sw_height - autoscroll_area - self.vp_css_margin:
             if not self.iv_auto_scroll_timer:
                 self.iv_auto_scroll_direction = Gtk.DirectionType.DOWN
                 self.iv_auto_scroll_timer = GObject.timeout_add(150,
@@ -1069,9 +1214,54 @@ class PdfArranger(Gtk.Application):
             GObject.source_remove(self.iv_auto_scroll_timer)
             self.iv_auto_scroll_timer = None
 
-        item = iconview.get_dest_item_at_pos(x, y)
-        if not item and self.is_between_items(iconview, x, y):
+        # Select move or copy dragAction
+        drag_move_posix = os.name == 'posix' and context.get_actions() & Gdk.DragAction.MOVE
+        drag_move_nt = os.name == 'nt' and not keyboard.is_pressed('control')
+        if drag_move_posix or drag_move_nt:
+            Gdk.drag_status(context, Gdk.DragAction.MOVE, etime)
+        else:
+            Gdk.drag_status(context, Gdk.DragAction.COPY, etime)
+
+        # By default 5 drag & drop positions are possible: into, left, right, above and below.
+        # We override default behaviour and only allow drag & drop to left or right.
+        # When drag location is a valid drop location True is returned.
+        model = iconview.get_model()
+        if len(model) == 0:
+            return True
+        cell_width, _cell_height = self.cellthmb.get_fixed_size()
+        row_distance = iconview.get_row_spacing() + 2 * iconview.get_item_padding()
+        column_distance = iconview.get_column_spacing() + 2 * iconview.get_item_padding()
+        search_positions = [('XY', x, y),
+                            ('Right', x + column_distance / 2, y),
+                            ('Left', x - column_distance / 2, y),
+                            ('Below', x, y + row_distance / 2),
+                            ('Above', x, y - row_distance / 2),
+                            ('Left-Above', x - column_distance, y - row_distance),
+                            ('Right-Far', x + cell_width, y),
+                            ('Left-Far', x - cell_width, y),
+                            ('Right-Below-Far', x + cell_width, y + row_distance),
+                            ('Left-Below-Far', x - cell_width, y + row_distance),
+                            ('Below-Far', x, y + row_distance)]
+        for search_pos, x_s, y_s in search_positions:
+            path = iconview.get_path_at_pos(x_s, y_s)
+            if path:
+                break
+        if search_pos in ['XY', 'Right', 'Left', 'Below', 'Above']:
+            self.drag_path = path
+            if path == iconview.get_path_at_pos(x_s + cell_width * 0.6, y_s):
+                self.drag_pos = Gtk.IconViewDropPosition.DROP_LEFT
+            elif path == iconview.get_path_at_pos(x_s - cell_width * 0.6, y_s):
+                self.drag_pos = Gtk.IconViewDropPosition.DROP_RIGHT
+        elif search_pos == 'Left-Above' and iconview.get_drag_dest_item()[0]:
+            return True
+        elif not path or (path == model[-1].path and x_s < x):
+            self.drag_path = model[-1].path
+            self.drag_pos = Gtk.IconViewDropPosition.DROP_RIGHT
+        else:
             iconview.stop_emission('drag_motion')
+            return False
+        iconview.set_drag_dest_item(self.drag_path, self.drag_pos)
+        return True
 
     def iv_dnd_leave_end(self, _widget, _context, _ignored=None):
         """Ends the auto-scroll during DND"""
@@ -1180,8 +1370,20 @@ class PdfArranger(Gtk.Application):
         """Handles received data by drag and drop in scrolledwindow"""
         if target_id == self.TEXT_URI_LIST:
             pageadder = PageAdder(self)
+            model = self.iconview.get_model()
+            ref_to = None
+            before = True
+            if len(model) > 0:
+                last_row = model[-1]
+                if self.drag_pos == Gtk.IconViewDropPosition.DROP_LEFT:
+                    ref_to = Gtk.TreeRowReference.new(model, self.drag_path)
+                elif self.drag_path != last_row.path:
+                    iter_next = model.iter_next(model.get_iter(self.drag_path))
+                    path_next = model.get_path(iter_next)
+                    ref_to = Gtk.TreeRowReference.new(model, path_next)
+            pageadder.move(ref_to, before)
             for uri in selection_data.get_uris():
-                filename = get_file_path_from_dnd_dropped_uri(uri)
+                filename = get_file_path_from_uri(uri)
                 pageadder.addpages(filename)
             pageadder.commit(select_added=False, add_to_undomanager=True)
 
@@ -1441,8 +1643,11 @@ class PdfArranger(Gtk.Application):
             self.set_export_file(None)
             self.set_unsaved(False)
 
-    def __update_num_pages(self, model, _path, _itr=None, _user_data=None):
-        self.uiXML.get_object("num_pages").set_text(str(len(model)))
+    def __update_num_pages(self, model, _path=None, _itr=None, _user_data=None):
+        num_pages = len(model)
+        self.uiXML.get_object("num_pages").set_text(str(num_pages))
+        for a in ["save", "save-as"]:
+            self.window.lookup_action(a).set_enabled(num_pages > 0)
 
     def error_message_dialog(self, msg, msg_type=Gtk.MessageType.ERROR):
         error_msg_dlg = Gtk.MessageDialog(flags=Gtk.DialogFlags.MODAL,
@@ -1454,19 +1659,47 @@ class PdfArranger(Gtk.Application):
             error_msg_dlg.destroy()
 
 
+class PDFDocError(Exception):
+    def __init__(self, message):
+        self.message = message
+
+
 class PDFDoc:
     """Class handling PDF documents"""
 
     def __init__(self, filename, tmp_dir):
-
         self.filename = os.path.abspath(filename)
         self.shortname = os.path.splitext(os.path.split(self.filename)[1])[0]
         self.mtime = os.path.getmtime(filename)
-        fd, self.copyname = tempfile.mkstemp(dir=tmp_dir)
-        os.close(fd)
-        shutil.copy(self.filename, self.copyname)
-        uri = pathlib.Path(self.copyname).as_uri()
-        self.document = Poppler.Document.new_from_file(uri, None)
+        filemime = mimetypes.guess_type(self.filename)[0]
+        if not filemime:
+            raise PDFDocError(_('Unknown file format'))
+        if filemime == 'application/pdf':
+            try:
+                fd, self.copyname = tempfile.mkstemp(dir=tmp_dir)
+                os.close(fd)
+                shutil.copy(self.filename, self.copyname)
+                uri = pathlib.Path(self.copyname).as_uri()
+                self.document = Poppler.Document.new_from_file(uri, None)
+            except GLib.Error as e:
+                raise PDFDocError(e.message + ': ' + filename)
+        elif filemime.split('/')[0] == 'image':
+            if not img2pdf:
+                raise PDFDocError(_('Image files are only supported with img2pdf'))
+            if mimetypes.guess_type(filename)[0] in img2pdf_supported_img:
+                try:
+                    fd, self.copyname = tempfile.mkstemp(dir=tmp_dir)
+                    os.close(fd)
+                    with open(self.copyname, 'wb') as f:
+                        f.write(img2pdf.convert(filename))
+                    uri = pathlib.Path(self.copyname).as_uri()
+                    self.document = Poppler.Document.new_from_file(uri, None)
+                except img2pdf.AlphaChannelError as e:
+                    raise PDFDocError(e)
+            else:
+                raise PDFDocError(_('Image format is not supported by img2pdf'))
+        else:
+            raise PDFDocError(_('File is neither pdf nor image'))
 
 
 class PageAdder(object):
@@ -1502,10 +1735,9 @@ class PageAdder(object):
         if not pdfdoc:
             try:
                 pdfdoc = PDFDoc(filename, self.app.tmp_dir)
-            except GLib.Error as e:
-                message = e.message + ': ' + filename
-                print(message, file=sys.stderr)
-                self.app.error_message_dialog(message)
+            except PDFDocError as e:
+                print(e.message, file=sys.stderr)
+                self.app.error_message_dialog(e.message)
                 return
             self.app.import_directory = os.path.split(filename)[0]
             self.app.export_directory = self.app.import_directory
